use sqlx::prelude::FromRow;

use crate::{posts::Post, BlogDb};

#[derive(Clone, FromRow, Debug)]
pub struct Draft {
    pub id: String,
    pub title: String,
    // Valid markdown
    pub content: String,
    // Separated by commas, no whitespace
    pub tags: String,
}

impl BlogDb {
    /// Create a draft from an existing post, returning a [Draft] either for the new draft or for an existing draft
    pub async fn edit_post<S>(&self, id: S) -> anyhow::Result<Draft>
    where
        S: AsRef<str>,
    {
        let draft_exists = sqlx::query_as::<_, Draft>("SELECT * FROM drafts WHERE id=?")
            .bind(id.as_ref())
            .fetch_one(&self.db)
            .await;

        if let Ok(draft) = draft_exists {
            Ok(draft)
        } else {
            let post_exists = sqlx::query_as::<_, Post>("SELECT * FROM posts WHERE id=?")
                .bind(id.as_ref())
                .fetch_one(&self.db)
                .await;
            if let Ok(post) = post_exists {
                let new_draft: Draft = sqlx::query_as("INSERT INTO drafts (id, title, tags, content) SELECT id, title, tags, content FROM posts WHERE id=? RETURNING *")
            .bind(post.id).fetch_one(&self.db).await?;
                Ok(new_draft)
            } else {
                Err(anyhow::Error::msg(
                    "Tried to edit a post that doesn't exist",
                ))
            }
        }
    }

    /// Get draft by id
    pub async fn get_draft<S>(&self, id: S) -> anyhow::Result<Draft>
    where
        S: AsRef<str>,
    {
        let draft: Draft = sqlx::query_as("SELECT * FROM drafts WHERE id=?")
            .bind(id.as_ref())
            .fetch_one(&self.db)
            .await?;
        Ok(draft)
    }

    /// Delete draft by id
    pub async fn delete_draft<S>(&self, id: S) -> anyhow::Result<()>
    where
        S: AsRef<str>,
    {
        sqlx::query_as::<_, Draft>("DELETE FROM drafts WHERE id=? RETURNING *")
            .bind(id.as_ref())
            .fetch_one(&self.db)
            .await?;
        Ok(())
    }

    /// Get all drafts
    pub async fn get_drafts(&self, session_id: i64) -> anyhow::Result<Vec<Draft>> {
        self.check_session(session_id).await?;
        let drafts: Vec<Draft> = sqlx::query_as("SELECT * FROM drafts")
            .fetch_all(&self.db)
            .await?;
        Ok(drafts)
    }

    /// Create a new draft with the given data. See [`edit_post`](Self::edit_post) to create a draft from an existing post
    pub async fn add_draft<S, T, R>(&self, title: S, tags: T, content: R) -> anyhow::Result<Draft>
    where
        S: AsRef<str>,
        T: AsRef<str>,
        R: AsRef<str>,
    {
        let draft: Draft =
            sqlx::query_as("REPLACE INTO drafts (title, tags, content) VALUES (?,?,?) RETURNING *")
                .bind(title.as_ref())
                .bind(tags.as_ref())
                .bind(content.as_ref())
                .fetch_one(&self.db)
                .await?;
        Ok(draft)
    }

    pub async fn update_draft_title<S, T>(&self, id: S, title: T) -> anyhow::Result<()>
    where
        S: AsRef<str>,
        T: AsRef<str>,
    {
        sqlx::query("UPDATE drafts SET (title) = ? WHERE id=?")
            .bind(title.as_ref())
            .bind(id.as_ref())
            .execute(&self.db)
            .await?;
        Ok(())
    }

    pub async fn update_draft_tags<S, T>(&self, id: S, tags: T) -> anyhow::Result<()>
    where
        S: AsRef<str>,
        T: AsRef<str>,
    {
        sqlx::query("UPDATE drafts SET (tags) = ? WHERE id=?")
            .bind(tags.as_ref())
            .bind(id.as_ref())
            .execute(&self.db)
            .await?;
        Ok(())
    }

    pub async fn update_draft_content<S, T>(&self, id: S, content: T) -> anyhow::Result<()>
    where
        S: AsRef<str>,
        T: AsRef<str>,
    {
        sqlx::query("UPDATE drafts SET (content) = ? WHERE id=?")
            .bind(content.as_ref())
            .bind(id.as_ref())
            .execute(&self.db)
            .await?;
        Ok(())
    }

    /// Publish draft by id, creating a new post if one with the id doesn't already exist, and updating the post if it does.
    ///
    /// Note that if a new post is created its id may be incremented, making the id passed into this function invalid.
    /// Always use the returned [`Post`] id for future actions.
    pub async fn publish_draft<S>(&self, id: S) -> anyhow::Result<Post>
    where
        S: AsRef<str>,
    {
        let update = sqlx::query_as::<_, Post>(
            "UPDATE posts
                SET (title, tags, content) = (SELECT title, tags, content FROM drafts WHERE id=?) 
                WHERE id=?
             RETURNING *",
        )
        .bind(id.as_ref())
        .fetch_one(&self.db)
        .await;

        if let Ok(post) = update {
            self.delete_draft(id).await?;
            Ok(post)
        } else {
            let new_post: Post = sqlx::query_as(
                "INSERT INTO posts (title, date, tags, content) 
                 SELECT title, DATE('now', 'localtime'), tags, content FROM drafts WHERE id=? 
                 RETURNING *",
            )
            .bind(id.as_ref())
            .fetch_one(&self.db)
            .await?;
            self.delete_draft(id).await?;
            Ok(new_post)
        }
    }
}

#[cfg(test)]
mod tests {
    use crate::util::tests::*;
    use sqlx::SqlitePool;

    #[sqlx::test]
    async fn drafts(pool: SqlitePool) -> anyhow::Result<()> {
        let (db, session_id, post) = init_and_add_post(pool).await?;

        // Test draft that updates existing post

        let id = post.id;
        db.edit_post(id).await?;

        let title = "this is an updated title";
        let content = "this is some updated content";
        let tags = "updated";

        db.update_draft_title(id, title).await?;
        db.update_draft_content(id, content).await?;
        db.update_draft_tags(id, tags).await?;
        db.publish_draft(id).await?;
        assert!(db.get_draft(id).await.is_err()); // Published drafts should be deleted

        let post = db.get_post(id).await?;
        assert_eq!(post.title, title.to_string());
        assert_eq!(post.content, content.to_string());
        assert_eq!(post.tags, tags.to_string());

        // Test new draft

        let title = "this is a new title";
        let content = "this is new content";
        let tags = "new";

        let draft = db.add_draft(title, tags, content).await?;
        assert!(db.get_draft(draft.id).await.is_ok());

        let post = db.publish_draft(draft.id).await?;
        assert!(db.get_post(post.id).await.is_ok());
        assert_eq!(post.title, title.to_string());
        assert_eq!(post.content, content.to_string());
        assert_eq!(post.tags, tags.to_string());

        Ok(())
    }
}
