mod assets;
mod drafts;
mod pages;
pub mod posts;
pub mod users;
mod util;

use posts::Post;
use util::*;

use std::fs::File;
use std::str::FromStr;

use sqlx::sqlite::SqliteJournalMode;
use sqlx::{
    sqlite::{SqliteConnectOptions, SqlitePoolOptions},
    SqlitePool,
};

/// A Sqlite database for all blog data.
///
/// It uses a file-backed db for posts, drafts, users, and assets, and an in-memory db for cached
/// web pages and session management.
#[derive(Clone)]
pub struct BlogDb {
    db: SqlitePool,
    memory: SqlitePool,
}

impl BlogDb {
    /// Create a new BlogDb with an initial user.
    pub async fn new<S>(username: S, password: String) -> anyhow::Result<Self>
    where
        S: AsRef<str>,
    {
        let main_db_name = dotenvy::var("DB_MAIN_PATH").unwrap_or("main.db".to_string());
        let cache_name = dotenvy::var("DB_CACHE_PATH").unwrap_or("cache.db".to_string());
        if File::open(&main_db_name).is_err() {
            File::create_new(&main_db_name)?;
        }
        if File::open(&cache_name).is_err() {
            File::create_new(&cache_name)?;
        }
        let db = SqlitePoolOptions::new()
            .connect_with(
                SqliteConnectOptions::from_str(&["sqlite://", &main_db_name].concat())?
                    .journal_mode(SqliteJournalMode::Wal),
            )
            .await?;
        let memory = SqlitePoolOptions::new()
            .min_connections(1)
            .connect_with(
                // #[cfg(not(debug_assertions))]
                // SqliteConnectOptions::from_str(MEMORY_URL)?.journal_mode(SqliteJournalMode::Memory),
                // #[cfg(debug_assertions)]
                SqliteConnectOptions::from_str(&["sqlite://", &cache_name].concat())?
                    .journal_mode(SqliteJournalMode::Wal),
            )
            .await?;
        Self::run_main_migrations(&db).await?;
        Self::run_memory_migrations(&memory).await?;
        Self::add_initial_user(&db, username.as_ref(), password).await?;
        Ok(Self { db, memory })
    }

    async fn run_main_migrations(db: &SqlitePool) -> anyhow::Result<()> {
        sqlx::migrate!("db/main/migrations").run(db).await?;
        Ok(())
    }

    async fn run_memory_migrations(cache: &SqlitePool) -> anyhow::Result<()> {
        sqlx::migrate!("db/memory/migrations").run(cache).await?;
        Ok(())
    }
}
