use sqlx::prelude::FromRow;

use crate::BlogDb;

#[derive(Clone, FromRow, Debug, Default)]
pub struct Post {
    pub id: String,
    pub title: String,
    pub date: String,
    // Valid markdown
    pub content: String,
    // Separated by commas, no whitespace
    pub tags: String,
}

impl BlogDb {
    /// Create a single post with the given data.
    ///
    /// This is mostly used for testing: most of the time you should use [`BlogDb::publish_draft`] to publish a post from a draft
    pub async fn add_post<S, T, R>(&self, title: S, tags: T, content: R) -> anyhow::Result<Post>
    where
        S: AsRef<str>,
        T: AsRef<str>,
        R: AsRef<str>,
    {
        let post: Post = sqlx::query_as(
            "REPLACE INTO posts (title, date, tags, content) VALUES (?, date(), ?, ?) RETURNING *",
        )
        .bind(title.as_ref())
        .bind(tags.as_ref())
        .bind(content.as_ref())
        .fetch_one(&self.db)
        .await?;
        Ok(post)
    }

    /// Get a single post by id
    pub async fn get_post<S>(&self, id: S) -> anyhow::Result<Post>
    where
        S: AsRef<str>,
    {
        let post: Post = sqlx::query_as("SELECT * FROM posts WHERE id=?")
            .bind(id.as_ref())
            .fetch_one(&self.db)
            .await?;
        Ok(post)
    }

    /// Get all posts
    pub async fn get_posts(&self) -> anyhow::Result<Vec<Post>> {
        let posts: Vec<Post> = sqlx::query_as("SELECT * FROM posts ORDER BY DATE(date) DESC")
            .fetch_all(&self.db)
            .await?;
        Ok(posts)
    }

    /// Get `count` number of posts
    pub async fn get_count_posts(&self, count: u32) -> anyhow::Result<Vec<Post>> {
        let posts: Vec<Post> =
            sqlx::query_as("SELECT * FROM posts LIMIT ? ORDER BY DATE(date) DESC")
                .bind(count)
                .fetch_all(&self.db)
                .await?;
        Ok(posts)
    }

    /// Delete post by id
    pub async fn delete_post<S>(&self, id: S) -> anyhow::Result<()>
    where
        S: AsRef<str>,
    {
        sqlx::query_as::<_, Post>("DELETE FROM posts WHERE id=? RETURNING *")
            .bind(id.as_ref())
            .fetch_one(&self.db)
            .await?;
        Ok(())
    }
}

#[cfg(test)]
mod tests {
    use crate::util::tests::*;
    use sqlx::SqlitePool;

    #[sqlx::test]
    async fn posts(pool: SqlitePool) -> anyhow::Result<()> {
        let (db, session_id) = get_init_db(pool).await?;

        let title = "ʕ·ᴥ·ʔ";
        let content = "ʕ·ᴥ·ʔ- hello there";
        let tags = "george";
        let post = db.add_post(title, tags, content).await?;

        let id = post.id;

        let post = db.get_post(id).await?;
        assert_eq!(post.title, title.to_string());
        assert_eq!(post.content, content.to_string());
        assert_eq!(post.tags, tags.to_string());

        db.delete_post(id).await?;
        let deleted = db.get_post(id).await.is_err();
        assert!(deleted);

        Ok(())
    }
}
