mod middleware;
pub(crate) mod routes;

use routes::{api::editor::Blocks, make_router};
use tantivy::{
    directory::MmapDirectory,
    doc,
    schema::{Schema, STORED, STRING, TEXT},
    Index, IndexWriter, TantivyDocument, Term,
};
use tower::Layer;
use tower_http::normalize_path::NormalizePathLayer;

use std::{
    env,
    fs::{self, File},
    io::{Read, Write},
    path::PathBuf,
    str::FromStr,
    sync::Arc,
    sync::Mutex,
};

use anyhow::anyhow;
use axum::{
    extract::{FromRef, Request},
    Router, ServiceExt,
};
use axum_extra::extract::cookie::Key;
use blogdb::BlogDb;
use glob::glob;

use crate::html;
include!(concat!(env!("OUT_DIR"), "/codegen.rs"));

const DB_URL: &str = "sqlite://evie.db";

pub struct Blog {
    app: Router,
}

impl Blog {
    pub async fn new() -> anyhow::Result<Self> {
        let state = BlogState::new().await?;
        let app = make_router(state).await;
        Ok(Self { app })
    }

    pub async fn serve(self, port: Option<u16>) {
        let port = port.unwrap_or(3000);
        let listener = tokio::net::TcpListener::bind(format!("127.0.0.1:{port}"))
            .await
            .unwrap();
        axum::serve(
            listener,
            ServiceExt::<Request>::into_make_service(
                NormalizePathLayer::trim_trailing_slash().layer(self.app),
            ),
        )
        .await
        .unwrap();
    }
}

#[derive(Clone)]
pub struct BlogState(Arc<BlogStateInner>);

impl BlogState {
    async fn new() -> anyhow::Result<Self> {
        Ok(BlogState(Arc::new(BlogStateInner::new().await?)))
    }

    fn db(&self) -> &BlogDb {
        &self.0.db
    }

    fn domain(&self) -> String {
        self.0.domain.clone()
    }

    fn index(&self) -> &Index {
        &self.0.index
    }

    fn index_add_document(&self, document: TantivyDocument) {
        let _ = self.0.index_writer.lock().unwrap().add_document(document);
    }

    fn index_delete(&self, term: Term) {
        let _ = self.0.index_writer.lock().unwrap().delete_term(term);
    }

    fn index_commit(&self) {
        let _ = self.0.index_writer.lock().unwrap().commit();
    }
}

// this impl tells `SignedCookieJar` how to access the key from our state
impl FromRef<BlogState> for Key {
    fn from_ref(state: &BlogState) -> Self {
        state.0.key.clone()
    }
}

struct BlogStateInner {
    db: BlogDb,
    key: Key,
    domain: String,
    index: Index,
    index_writer: Mutex<IndexWriter>,
}

impl BlogStateInner {
    async fn new() -> anyhow::Result<Self> {
        let username = env!("INIT_USER_NAME").to_string();
        let password = env!("INIT_USER_PASSWORD").to_string();
        let domain = option_env!("DOMAIN").unwrap_or("localhost").to_string();
        let db = BlogDb::new(&username, password.clone()).await?;
        Self::add_initial_assets(&db).await?;
        Self::add_initial_pages(&db).await?;
        #[cfg(debug_assertions)]
        {
            Self::add_posts(&db).await?;
            Self::add_drafts(&username, password, &db).await?;
        }
        let key = Self::key()?;
        let (index, index_writer) = Self::generate_index(&db).await?;
        Ok(Self {
            db,
            key,
            domain,
            index,
            index_writer: Mutex::new(index_writer),
        })
    }

    async fn add_initial_pages(db: &BlogDb) -> anyhow::Result<()> {
        let pages = vec![
            ("index".to_string(), html::home_page(db).await),
            ("about".to_string(), html::about_page(db).await),
        ];
        db.add_pages(pages).await?;
        Ok(())
    }

    async fn add_initial_assets(db: &BlogDb) -> anyhow::Result<()> {
        include!(concat!(env!("OUT_DIR"), "/add_assets.rs"));
        Ok(())
    }

    const INDEX_PATH: &'static str = "./index";
    const MEMORY_BUDGET: usize = 50_000_000;

    async fn generate_index(db: &BlogDb) -> anyhow::Result<(Index, IndexWriter)> {
        let mut schema_builder = Schema::builder();
        let id = schema_builder.add_text_field("id", STRING | STORED);
        let title = schema_builder.add_text_field("title", TEXT | STORED);
        let body = schema_builder.add_text_field("body", TEXT | STORED);
        let schema = schema_builder.build();
        let _ = fs::remove_dir_all(Self::INDEX_PATH);
        let _ = fs::create_dir_all(Self::INDEX_PATH);
        let dir = MmapDirectory::open(Self::INDEX_PATH).unwrap();
        let index = Index::open_or_create(dir, schema.clone())?;
        let mut index_writer: IndexWriter = index.writer(Self::MEMORY_BUDGET)?;
        let posts = db.get_posts().await?;
        for post in posts.into_iter() {
            let content: Blocks = serde_json::from_str(&post.content).unwrap_or_default();
            index_writer.add_document(
                doc!(id => post.id, title => post.title, body => content.to_plaintext()),
            )?;
        }
        index_writer.commit()?;
        Ok((index, index_writer))
    }

    #[cfg(not(debug_assertions))]
    fn key() -> anyhow::Result<Key> {
        let key = Key::try_generate().ok_or(anyhow!("Couldn't generate key for cookie signing"))?;
        Ok(key)
    }

    #[cfg(debug_assertions)]
    fn key() -> anyhow::Result<Key> {
        let key_path = PathBuf::from_str("./key")?;
        let key_file = File::options().read(true).write(true).open(&key_path);
        let key = match key_file {
            Ok(mut file) => {
                let mut bytes = vec![];
                file.read_to_end(&mut bytes)?;
                match Key::try_from(bytes.as_slice()) {
                    Ok(key) => key,
                    Err(error) => {
                        println!("need to make a new key, couldn't get from file cause {error:?}");
                        let key = Key::try_generate()
                            .ok_or(anyhow!("Couldn't generate key for cookie signing"))?;
                        file.write_all(key.master())?;
                        println!("made one");
                        key
                    }
                }
            }
            Err(error) => {
                println!("need to make a new key, couldn't open the key file cause {error:?}");
                let key = Key::try_generate()
                    .ok_or(anyhow!("Couldn't generate key for cookie signing"))?;

                let mut key_file = File::create(&key_path)?;
                key_file.write_all(key.master())?;
                println!("made one");
                key
            }
        };
        Ok(key)
    }

    #[cfg(debug_assertions)]
    async fn add_posts(db: &BlogDb) -> anyhow::Result<()> {
        let content = r#"
[
  {
    "id": "26lQiP26Dd",
    "type": "header",
    "data": {
      "text": "Example Post",
      "level": 1
    }
  },
  {
    "id": "PnWeuvNW6O",
    "type": "header",
    "data": {
      "text": "This is a subheading",
      "level": 2
    }
  },
  {
    "id": "xSV-1St5Nh",
    "type": "warning",
    "data": {
      "title": "This is a warning...",
      "message": ""
    }
  },
  {
    "id": "rfnEtkKIjP",
    "type": "quote",
    "data": {
      "text": "this is a quote<br>",
      "caption": "",
      "alignment": "left"
    }
  },
  {
    "id": "6lQRMa2guf",
    "type": "paragraph",
    "data": {
      "text": "This is some content<br>"
    }
  }
]
            "#;
        if db.get_posts().await?.is_empty() {
            db.add_post("Example Post", "", content).await?;
        }
        Ok(())
    }

    #[cfg(debug_assertions)]
    async fn add_drafts(username: &str, password: String, db: &BlogDb) -> anyhow::Result<()> {
        let content = r#"
[
  {
    "id": "26lQiP26Dd",
    "type": "header",
    "data": {
      "text": "Example Draft",
      "level": 1
    }
  },
  {
    "id": "PnWeuvNW6O",
    "type": "header",
    "data": {
      "text": "This is a subheading",
      "level": 2
    }
  },
  {
    "id": "xSV-1St5Nh",
    "type": "warning",
    "data": {
      "title": "This is a warning...",
      "message": ""
    }
  },
  {
    "id": "rfnEtkKIjP",
    "type": "quote",
    "data": {
      "text": "this is a quote<br>",
      "caption": "",
      "alignment": "left"
    }
  },
  {
    "id": "6lQRMa2guf",
    "type": "paragraph",
    "data": {
      "text": "This is some content<br>"
    }
  }
]
            "#;
        let session_id = db.new_session(username, password, "+1 year").await?;
        if db.get_drafts(session_id).await?.is_empty() {
            db.add_draft("Example Draft", "", content).await?;
        }
        db.end_session(session_id).await?;
        Ok(())
    }
}
