use axum::{
    extract::State,
    http::{header::LOCATION, HeaderMap, HeaderValue, StatusCode},
    response::{Html, IntoResponse, Response},
    routing::post,
    Form, Router,
};
use axum_extra::extract::{cookie::Cookie, PrivateCookieJar};
use blogdb::BlogDb;
use serde::Deserialize;
use serde_json::to_string;

use crate::{html, BlogState};

#[derive(Deserialize)]
struct LoginCredentials {
    username: Option<String>,
    password: Option<String>,
    remember: Option<String>,
    next: Option<String>,
}

pub(super) fn auth(State(state): State<BlogState>) -> Router {
    Router::new()
        .route("/login", post(login))
        .route("/logout", post(logout))
        .with_state(state)
}

async fn login(
    mut jar: PrivateCookieJar,
    State(state): State<BlogState>,
    Form(body): Form<LoginCredentials>,
) -> Response {
    let db = &state.db();
    let next = match body.next.unwrap_or("/".to_string()).as_str() {
        "/" => "/".to_string(),
        next => ["/", next].concat(),
    };
    let username = match body.username {
        Some(username) => username,
        None => return bad_login(&next).await,
    };
    let password = match body.password {
        Some(password) => password,
        None => return bad_login(&next).await,
    };
    let expiration = if body.remember.is_some() {
        "+1 year"
    } else {
        "+0 day"
    };
    match db.new_session(username, password, expiration).await {
        Ok(id) => {
            jar = jar.add(
                Cookie::build(("id", id.to_string()))
                    .domain(state.domain())
                    .secure(true)
                    .http_only(true)
                    .path("/"),
            );
            let mut headers = HeaderMap::new();
            headers.insert(LOCATION, HeaderValue::from_str(&next).unwrap());
            (StatusCode::FOUND, headers, jar).into_response()
        }
        Err(_) => bad_login(&next).await,
    }
}

async fn bad_login(next: &str) -> Response {
    (
        StatusCode::UNAUTHORIZED,
        Html(html::login_status(next, false)),
    )
        .into_response()
}

async fn logout(jar: PrivateCookieJar, State(state): State<BlogState>) -> Response {
    let db = &state.db();
    let session_id: i64 = if let Some(cookie) = jar.get("id") {
        match cookie.value().parse() {
            Ok(id) => id,
            Err(_) => return StatusCode::BAD_REQUEST.into_response(),
        }
    } else {
        return StatusCode::BAD_REQUEST.into_response();
    };
    match db.end_session(session_id).await {
        Ok(_) => {
            let jar = jar.remove("id");
            let mut headers = HeaderMap::new();
            headers.insert(LOCATION, HeaderValue::from_str("/").unwrap());
            (StatusCode::FOUND, headers, jar).into_response()
        }
        Err(_) => StatusCode::BAD_REQUEST.into_response(),
    }
}
