use axum::{
    extract::{Json, Path, State},
    http::StatusCode,
    response::{IntoResponse, Response},
    routing::post,
    Router,
};
use axum_extra::extract::PrivateCookieJar;
use blogdb::BlogDb;
use serde::{Deserialize, Serialize};

use crate::{
    html::{self, about_page, home_page, remove_el},
    BlogState,
};

#[derive(Debug, Deserialize, Serialize)]
#[serde(tag = "type", content = "data")]
#[allow(non_camel_case_types)]
pub(crate) enum Block {
    paragraph {
        text: String,
    },
    header {
        text: String,
        level: usize,
    },
    list {
        style: ListStyle,
        items: Vec<String>,
    },
    warning {
        title: String,
    },
    quote {
        text: String,
        caption: String,
    },
    embed {
        service: String,
        source: String,
        embed: String,
        width: usize,
        height: usize,
    },
    image {
        file: ImageSource,
        #[serde(default)]
        caption: String,
    },
    delimiter {},
}

#[derive(Debug, Deserialize, Serialize)]
#[allow(non_camel_case_types)]
pub struct ImageSource {
    url: String,
}

#[derive(Debug, Deserialize, Serialize)]
#[allow(non_camel_case_types)]
pub enum ListStyle {
    ordered,
    unordered,
}

impl Block {
    fn to_html(&self) -> String {
        let text = match self {
            Block::paragraph { text } => ["<p>", text, "</p>"].concat(),
            Block::header { text, level } => {
                let level = level.to_string();
                ["<h", &level, ">", text, "</h", &level, ">"].concat()
            }
            Block::list { style, items } => match style {
                ListStyle::ordered => ["<ol><li>", &items.join("</li><li>"), "</li></ol>"].concat(),
                ListStyle::unordered => {
                    ["<ul><li>", &items.join("</li><li>"), "</li></ul>"].concat()
                }
            },
            Block::warning { title } => ["<aside>", title, "</aside>"].concat(),
            Block::quote { text, caption } => {
                if caption.is_empty() {
                 ["<figure><blockquote>“", text, "”</blockquote></figure>"].concat()
                } else {
                 ["<figure><blockquote>“", text, "”</blockquote><figcaption>— ", caption,"</figcaption></figure>"].concat()
                }
            },
            Block::embed {
                embed,
                width,
                height,
                ..
            } => [
                "<iframe width=\"",
                &width.to_string(),
                "\" height=\"",
                &height.to_string(),
                "\" src=\"",
                embed,
                "\"></iframe>",
            ]
            .concat(),
            Block::image { file, caption } => {
                ["<img src=\"", &file.url, "\" alt=\"", &caption.replace(r#"""#, "&quot;").replace("<br>", ""), "\"/>"].concat()
            }
            Block::delimiter {} => "<div style=\"inline-size: 100%; block-size: 1px; background: var(--color-text)\"></div>".to_string(),

        };
        html::animate_anchors(html::remove_el(&text, "br"))
    }

    fn to_plaintext(&self) -> String {
        let text = match self {
            Block::paragraph { text } => text,
            Block::header { text, level: _ } => text,
            Block::list { style: _, items } => &items.join("\n"),
            Block::warning { title } => title,
            Block::quote { text, caption } => &[text, caption.as_str()].concat(),
            _ => &"".to_string(),
        };
        let text = [text, "\n"].concat();
        remove_el(&text, "br")
    }
}

pub(super) fn editor(State(state): State<BlogState>) -> Router {
    Router::new()
        .route("/update/:id", post(update))
        .route("/update/about", post(update_about))
        .route("/update/home", post(update_homepage))
        .with_state(state)
}

pub(crate) async fn check_id(
    jar: PrivateCookieJar,
    db: &BlogDb,
) -> Result<(i64, String), Response> {
    match jar.get("id") {
        Some(id) => match id.value_trimmed().parse::<i64>() {
            Ok(id) => match db.check_session(id).await {
                Ok(user) => Ok((id, user)),
                Err(_) => Err(StatusCode::UNAUTHORIZED.into_response()),
            },
            Err(_) => Err(StatusCode::BAD_REQUEST.into_response()),
        },
        None => Err(StatusCode::UNAUTHORIZED.into_response()),
    }
}

async fn update(
    jar: PrivateCookieJar,
    Path(id): Path<String>,
    State(state): State<BlogState>,
    Json(content): Json<Blocks>,
) -> Response {
    let db = state.db();

    if let Err(response) = check_id(jar, db).await {
        return response;
    }

    let title = content.title();
    let _ = db.update_draft_title(&id, title).await;

    let _ = db
        .update_draft_content(&id, serde_json::to_string(&content).unwrap())
        .await;

    StatusCode::OK.into_response()
}

async fn update_about(
    jar: PrivateCookieJar,
    State(state): State<BlogState>,
    Json(content): Json<Blocks>,
) -> Response {
    let db = state.db();

    let (id, user) = match check_id(jar, db).await {
        Ok(info) => info,
        Err(response) => return response,
    };

    let _ = db
        .update_user_about(&user, serde_json::to_string(&content).unwrap(), id)
        .await;

    let _ = db.update_page("about", about_page(db).await).await;
    StatusCode::OK.into_response()
}
async fn update_homepage(
    jar: PrivateCookieJar,
    State(state): State<BlogState>,
    Json(content): Json<Blocks>,
) -> Response {
    let db = state.db();

    let (id, user) = match check_id(jar, db).await {
        Ok(info) => info,
        Err(response) => return response,
    };

    let _ = db
        .update_user_homepage(&user, serde_json::to_string(&content).unwrap(), id)
        .await;

    let _ = db.update_page("index", home_page(db).await).await;
    StatusCode::OK.into_response()
}

#[derive(Debug, Deserialize, Serialize, Default)]
pub(crate) struct Blocks(Vec<Block>);

impl Blocks {
    pub(crate) fn to_html(&self) -> String {
        let mut html = String::new();
        for block in self.0.iter() {
            html.push_str(&block.to_html());
        }
        html
    }
    pub(crate) fn to_plaintext(&self) -> String {
        let mut text = String::new();
        for block in self.0.iter() {
            text.push_str(&block.to_plaintext());
        }
        text
    }

    pub(crate) fn first(&self) -> Option<&Block> {
        self.0.first()
    }

    pub(crate) fn title(&self) -> &str {
        if let Some(Block::header { text, level: 1 }) = self.first() {
            text
        } else {
            "Untitled"
        }
    }

    pub(crate) fn desc(&self) -> String {
        html::sanitize(
            self.0
                .iter()
                .find(|block| match block {
                    Block::paragraph { text: _ } => true,
                    Block::header { text: _, level } => *level > 1,
                    Block::warning { title: _ } => true,
                    Block::quote {
                        text: _,
                        caption: _,
                    } => true,
                    _ => false,
                })
                .map(|block| match block {
                    Block::paragraph { text } => text,
                    Block::header { text, level: _ } => text,
                    Block::warning { title } => title,
                    Block::quote { text, caption: _ } => text,
                    _ => "...",
                })
                .unwrap_or("No description"),
        )
    }
    pub(crate) fn image(&self) -> Option<&str> {
        self.0
            .iter()
            .find(|block| matches!(block, Block::image { .. }))
            .map(|block| match block {
                Block::image { file, .. } => Some(file.url.as_str()),
                _ => None,
            })
            .unwrap_or_default()
    }
}
