use axum::{
    extract::{Path, State},
    handler::HandlerWithoutStateExt,
    http::{header::LOCATION, HeaderMap, HeaderValue, StatusCode},
    response::{IntoResponse, Response},
    routing::{get, post},
    Router,
};

use axum_extra::extract::{Form, PrivateCookieJar, Query};

use serde::Deserialize;
use tantivy::{
    doc,
    schema::{Schema, STORED, TEXT},
    Term,
};

use crate::BlogState;

mod auth;
pub(crate) mod editor;
use auth::auth;
use editor::{check_id, editor, Blocks};

pub(super) fn api(state: BlogState) -> Router {
    Router::new()
        .route("/posts/:path", post(posts))
        .route("/user/update", post(user_update))
        .route("/drafts/:path", post(drafts))
        .route("/search", get(search))
        .with_state(state.clone())
        .nest("/auth", auth(State(state.clone())))
        .nest("/editor", editor(State(state.clone())))
}

#[derive(Debug, Deserialize)]
struct Entries {
    #[serde(default)]
    item: Vec<String>,
}

#[derive(Debug, Deserialize)]
#[allow(non_camel_case_types)]
enum PostsEndpoints {
    delete,
    unpublish,
}

async fn posts(
    Path(path): Path<PostsEndpoints>,
    State(state): State<BlogState>,
    Form(data): Form<Entries>,
) -> Response {
    let db = state.db();
    let mut schema_builder = Schema::builder();
    let id = schema_builder.add_text_field("id", TEXT | STORED);
    match path {
        PostsEndpoints::delete => {
            for post_id in data.item.iter() {
                let _ = db.delete_post(post_id).await;
                let term = Term::from_field_text(id, post_id);
                state.index_delete(term);
            }
        }
        PostsEndpoints::unpublish => {
            for post_id in data.item.iter() {
                let _ = db.edit_post(post_id).await;
                let _ = db.delete_post(post_id).await;
                let term = Term::from_field_text(id, post_id);
                state.index_delete(term);
            }
        }
    }
    state.index_commit();
    let mut headers = HeaderMap::new();
    headers.insert(LOCATION, HeaderValue::from_str("/admin").unwrap());
    (StatusCode::FOUND, headers).into_response()
}
#[derive(Debug, Deserialize)]
#[allow(non_camel_case_types)]
enum DraftsEndpoints {
    delete,
    publish,
}

async fn drafts(
    Path(path): Path<DraftsEndpoints>,
    State(state): State<BlogState>,
    Form(data): Form<Entries>,
) -> Response {
    let db = state.db();
    match path {
        DraftsEndpoints::delete => {
            for id in data.item.iter() {
                let _result = db.delete_draft(id).await;
            }
        }
        DraftsEndpoints::publish => {
            for id in data.item.iter() {
                match db.publish_draft(id).await {
                    Ok(post) => {
                        let mut schema_builder = Schema::builder();
                        let id = schema_builder.add_text_field("id", TEXT | STORED);
                        let body = schema_builder.add_text_field("body", TEXT | STORED);
                        let title = schema_builder.add_text_field("title", TEXT | STORED);
                        let content: Blocks =
                            serde_json::from_str(&post.content).unwrap_or_default();
                        state.index_add_document(doc!(id => post.id, body => content.to_plaintext(), title => post.title));
                    }
                    Err(_) => continue,
                };
            }
            state.index_commit();
        }
    }
    let mut headers = HeaderMap::new();
    headers.insert(LOCATION, HeaderValue::from_str("/admin").unwrap());
    (StatusCode::FOUND, headers).into_response()
}

#[derive(Debug, Deserialize)]
struct SearchQuery {
    #[serde(default)]
    query: Vec<String>,
}

async fn search(Query(query): Query<SearchQuery>) -> Response {
    let query = query
        .query
        .join(" ")
        .split(" ")
        .collect::<Vec<&str>>()
        .join("-");
    let url = ["/search/", &query].concat();

    let mut headers = HeaderMap::new();
    headers.insert(LOCATION, HeaderValue::from_str(&url).unwrap());
    (StatusCode::FOUND, headers).into_response()
}

#[derive(Deserialize)]
struct UpdateUserInfo {
    username: String,
    password: String,
}

async fn user_update(
    jar: PrivateCookieJar,
    State(state): State<BlogState>,
    Form(info): Form<UpdateUserInfo>,
) -> Response {
    let db = state.db();
    let (id, user) = match check_id(jar, db).await {
        Ok(creds) => creds,
        Err(resp) => return resp,
    };
    if info.username != user {
        let _ = db.update_user_name(&user, &info.username, id).await;
        let _ = db.update_password(user, info.password, id).await;
    } else {
        let _ = db.update_password(user, info.password, id).await;
    }
    let mut headers = HeaderMap::new();
    headers.insert(LOCATION, HeaderValue::from_str("/admin").unwrap());
    (StatusCode::FOUND, headers).into_response()
}
