pub(crate) mod api;
mod login;
use std::io::Cursor;

use api::{api, editor::check_id};
use axum_extra::{extract::PrivateCookieJar, response::Html};
use login::login;

use axum::{
    body::Body,
    extract::{rejection::PathRejection, DefaultBodyLimit, Multipart, Path, State},
    http::{
        header::{CACHE_CONTROL, CONTENT_TYPE, EXPIRES, LOCATION, PRAGMA},
        HeaderMap, HeaderValue, StatusCode,
    },
    middleware::from_fn_with_state,
    response::{IntoResponse, Response},
    routing::{get, post},
    Json, Router,
};
use serde::{Deserialize, Serialize};
use tantivy::{
    collector::TopDocs,
    query::QueryParser,
    schema::{Schema, STRING, TEXT},
    DocAddress, Document, Score, TantivyDocument,
};
use tokio_util::io::ReaderStream;
use tower_http::{limit::RequestBodyLimitLayer, trace::TraceLayer};
use tracing_subscriber::{layer::SubscriberExt, util::SubscriberInitExt};

use crate::{
    html::{
        self, about_editor, admin_page, draft_editor, homepage_editor, search_page,
        search_page_no_query,
    },
    BlogState,
};

use super::middleware::logged_in;

pub(super) async fn make_router(state: BlogState) -> Router {
    #[cfg(debug_assertions)]
    tracing_subscriber::registry()
        .with(
            tracing_subscriber::EnvFilter::try_from_default_env().unwrap_or_else(|_| {
                // axum logs rejections from built-in extractors with the `axum::rejection`
                // target, at `TRACE` level. `axum::rejection=trace` enables showing those events
                format!(
                    "{}=debug,tower_http=debug,axum::rejection=trace",
                    env!("CARGO_CRATE_NAME")
                )
                .into()
            }),
        )
        .with(tracing_subscriber::fmt::layer())
        .init();
    let router = Router::new()
        .merge(protected_routes(state.clone()).await)
        .merge(blog(State(state.clone())))
        .route("/assets/*asset", get(asset).with_state(state.clone()))
        .nest("/api", api(state.clone()))
        .nest("/login", login(State(state.clone())))
        .route("/", get(pages).with_state(state.clone()))
        .route("/*path", get(pages).with_state(state.clone()))
        .route("/search", get(search_empty))
        .route("/search/*query", get(search).with_state(state.clone()))
        .fallback(get(handle_404))
        .layer(DefaultBodyLimit::disable())
        .layer(RequestBodyLimitLayer::new(250 * 1024 * 1024));
    #[cfg(debug_assertions)]
    let router = router.layer(TraceLayer::new_for_http());
    router
}

async fn protected_routes(state: BlogState) -> Router {
    Router::new()
        .route("/admin", get(admin).with_state(state.clone()))
        .route("/editor/:id", get(editor).with_state(state.clone()))
        .route(
            "/editor/uploadFile",
            post(editor_upload_file).with_state(state.clone()),
        )
        .route(
            "/editor/fetchUrl",
            post(editor_fetch_url).with_state(state.clone()),
        )
        .route_layer(from_fn_with_state(state.clone(), logged_in))
}

async fn pages(
    path: Result<Path<String>, PathRejection>,
    State(state): State<BlogState>,
) -> Response {
    match path {
        Ok(path) => match state.db().get_page(path.0).await {
            Ok(page) => Html(page).into_response(),
            Err(_) => four_oh_four(),
        },
        Err(error) => match error {
            PathRejection::FailedToDeserializePathParams(_) => {
                match state.db().get_page("/").await {
                    Ok(page) => Html(page).into_response(),
                    Err(_) => four_oh_four(),
                }
            }
            _ => four_oh_four(),
        },
    }
}

pub(super) fn blog(State(state): State<BlogState>) -> Router {
    Router::new()
        .route("/blog", get(blog_root))
        .route("/blog/*path", get(blog_post))
        .with_state(state)
}

async fn blog_root(State(state): State<BlogState>) -> Response {
    let db = state.db();
    let page = html::blog_roll(db).await;
    Html(page).into_response()
}

async fn blog_post(Path(id): Path<String>, State(state): State<BlogState>) -> Response {
    let db = state.db();
    match db.get_post(id).await {
        Ok(post) => {
            let html = html::blog_page(post.id, db).await;
            Html(html).into_response()
        }
        Err(_) => four_oh_four(),
    }
}

pub(crate) fn four_oh_four() -> Response {
    (
        StatusCode::NOT_FOUND,
        Html(html::make_404("Page not found!")),
    )
        .into_response()
}

pub(crate) async fn editor(
    jar: PrivateCookieJar,
    Path(id): Path<String>,
    State(state): State<BlogState>,
) -> Response {
    let db = state.db();
    match id.as_str() {
        "new" => match db.add_draft("Untitled", "", "").await {
            Ok(draft) => {
                let mut headers = HeaderMap::new();
                headers.insert(
                    LOCATION,
                    HeaderValue::from_str(&format!("/editor/{}", draft.id)).unwrap(),
                );
                (StatusCode::FOUND, headers).into_response()
            }
            Err(_) => {
                let mut headers = HeaderMap::new();
                headers.insert(LOCATION, HeaderValue::from_str("/admin").unwrap());
                (StatusCode::FOUND, headers).into_response()
            }
        },
        "about" => {
            let (_id, user) = match check_id(jar, db).await {
                Ok(info) => info,
                Err(response) => return response,
            };
            let about_content = db.get_user_info(user).await.unwrap().about;
            let page = about_editor(&about_content);
            Html(page).into_response()
        }
        "home" => {
            let (_id, user) = match check_id(jar, db).await {
                Ok(info) => info,
                Err(response) => return response,
            };
            let home_content = db.get_user_info(user).await.unwrap().home;
            let page = homepage_editor(&home_content);
            Html(page).into_response()
        }
        _ => {
            let draft = match db.edit_post(id).await {
                Ok(draft) => draft,
                Err(_) => {
                    let mut headers = HeaderMap::new();
                    headers.insert(LOCATION, HeaderValue::from_str("/editor/new").unwrap());
                    return (StatusCode::FOUND, headers).into_response();
                }
            };
            let page = draft_editor(&draft.content, &draft.id.to_string());
            Html(page).into_response()
        }
    }
}

#[derive(Serialize)]
struct UploadResult {
    success: usize,
    file: Option<UrlUpload>,
}
impl UploadResult {
    fn success(url: String) -> Self {
        Self {
            success: 1,
            file: Some(UrlUpload { url }),
        }
    }
    fn failure() -> Self {
        Self {
            success: 0,
            file: None,
        }
    }
}

pub(crate) async fn editor_upload_file(
    State(state): State<BlogState>,
    mut fields: Multipart,
) -> Response {
    let db = state.db();
    let field = match fields.next_field().await {
        Ok(option) => match option {
            Some(field) => field,
            None => {
                return (StatusCode::BAD_REQUEST, Json(UploadResult::failure())).into_response()
            }
        },
        Err(_) => return (StatusCode::BAD_REQUEST, Json(UploadResult::failure())).into_response(),
    };
    let name = match field.file_name() {
        Some(name) => name.to_string(),
        None => "".to_string(),
    };
    let data = match field.bytes().await {
        Ok(data) => data,
        Err(_) => return (StatusCode::BAD_REQUEST, Json(UploadResult::failure())).into_response(),
    };

    match db.add_asset(name, data).await {
        Ok(asset) => {
            let slug = ["/assets/", &asset.slug].concat();
            Json(UploadResult::success(slug)).into_response()
        }
        Err(_error) => (
            StatusCode::INTERNAL_SERVER_ERROR,
            Json(UploadResult::failure()),
        )
            .into_response(),
    }
}

#[derive(Deserialize, Serialize)]
pub struct UrlUpload {
    url: String,
}
pub(crate) async fn editor_fetch_url(
    // jar: PrivateCookieJar,
    State(state): State<BlogState>,
    Json(UrlUpload { url: _ }): Json<UrlUpload>,
) -> Response {
    let _db = state.db();
    (StatusCode::INTERNAL_SERVER_ERROR, Json(r#"{"success": 0}"#)).into_response()
}

async fn asset(Path(slug): Path<String>, State(state): State<BlogState>) -> Response {
    let db = state.db();
    match db.get_asset(slug).await {
        Ok(asset) => {
            async {
                // i still don't quite know how to do async rust
                let stream = ReaderStream::new(Cursor::new(asset.data));
                let body = Body::from_stream(stream);
                let mut headers = HeaderMap::new();
                headers.append(CONTENT_TYPE, HeaderValue::from_str(&asset.mime).unwrap());
                (headers, body).into_response()
            }
            .await
        }
        Err(_) => StatusCode::NOT_FOUND.into_response(),
    }
}

async fn handle_404() -> Response {
    four_oh_four()
}

async fn admin(jar: PrivateCookieJar, State(state): State<BlogState>) -> Response {
    let db = state.db();
    // this can be unwrapped because our login middleware
    // ensures the user has a valid id
    let session_id = jar
        .get("id")
        .unwrap()
        .value_trimmed()
        .parse()
        .unwrap_or_default();
    let page = admin_page(session_id, db).await;
    let mut headers = HeaderMap::new();
    headers.append(
        CACHE_CONTROL,
        HeaderValue::from_str("no-cache, no-store, must-revalidate").unwrap(),
    );
    headers.append(PRAGMA, HeaderValue::from_str("no-cache").unwrap());
    headers.append(EXPIRES, HeaderValue::from_str("0").unwrap());
    (headers, Html(page)).into_response()
}

#[derive(Serialize, Deserialize, Default, Debug)]
pub(crate) struct SearchResponse(Vec<SearchResponseEntry>);
impl SearchResponse {
    pub(crate) fn is_empty(&self) -> bool {
        self.0.is_empty()
    }
}

impl IntoIterator for SearchResponse {
    type IntoIter = <Vec<SearchResponseEntry> as IntoIterator>::IntoIter;
    type Item = SearchResponseEntry;

    fn into_iter(self) -> Self::IntoIter {
        self.0.into_iter()
    }
}

#[derive(Serialize, Deserialize, Debug)]
pub(crate) struct SearchResponseEntry {
    // idk why tantivy makes these fields sequences
    body: Vec<String>,
    pub(crate) id: Vec<String>,
    title: Vec<String>,
}

async fn search_empty() -> Response {
    Html(search_page_no_query()).into_response()
}

async fn search(Path(query): Path<String>, State(state): State<BlogState>) -> Response {
    let query = query.split("-").collect::<Vec<&str>>().join(" ");

    let closure_query = query.clone();

    let closure_state = state.clone();
    let response = tokio::task::spawn_blocking(move || {
        let index = closure_state.index();
        let reader = index.reader().unwrap();
        let searcher = reader.searcher();
        let mut schema_builder = Schema::builder();
        let (id, title, body) = (
            schema_builder.add_text_field("id", STRING),
            schema_builder.add_text_field("title", TEXT),
            schema_builder.add_text_field("body", TEXT),
        );
        let schema = schema_builder.build();
        let query_parser = QueryParser::for_index(index, vec![id, title, body]);
        let query = query_parser.parse_query(&closure_query).unwrap();
        let top_docs: Vec<(Score, DocAddress)> =
            searcher.search(&query, &TopDocs::with_limit(100)).unwrap();
        let mut json = String::new();
        json.push('[');
        let mut docs_iter = top_docs.iter().peekable();
        while let Some((_score, doc_address)) = docs_iter.next() {
            let retrieved = searcher
                .doc::<TantivyDocument>(*doc_address)
                .unwrap_or_default();
            json.push_str(&retrieved.to_json(&schema));
            if docs_iter.peek().is_some() {
                json.push(',');
            }
        }
        json.push(']');
        // TODO: Deserialize directly to SearchResponseEntry
        let result = serde_json::from_str::<SearchResponse>(&json);
        result.unwrap_or_default()
    })
    .await
    .unwrap_or_default();

    let db = state.db();
    let html = search_page(query, response, db).await;
    Html(html).into_response()
}
